/* code.c */
#include <stdlib.h>
#include <ctype.h>
#include "err.h"
#include "code.h"
#include "server.h"

static int  hexval(char c) {
  if ('0'<=c && c<='9') return c-'0';
  if ('a'<=c && c<='f') return c-'a'+10;
  if ('A'<=c && c<='F') return c-'A'+10;
  return 0;
}
static char hex2(char *str) { return hexval(str[0])<<4 | hexval(str[1]); }

/* decode url string (returns malloc'ed string) */
char *decode(char *url) {
  char *cp;
  char *retval = malloc(1+strlen(url)); CHK_NULL(retval);
  for (cp = retval; *url!='\0' && *url!='?' && 
	            *url!=' '  && *url!='#'; cp++, url++)
    if (url[0]!=',' || url[1]=='\0' || url[2]=='\0') *cp=*url;
    else { *cp=hex2(url+1); url+=2; }
  /* copy post-url bit */
  while (*url!='\0') *cp++=*url++;
  /* null-terminate */
  *cp='\0';
  return retval;
}

static void expand(char *dst, char *src) {
  static char *hexdigit="0123456789ABCDEF";
  while (*src!='\0' && *src!='?' && *src!=' ' && *src!='#') {
    if (isalnum(*src) || *src=='/' || *src=='.' || *src=='_' || 
	*src=='-' || *src=='$')
      *dst++=*src++;
    else {
      *dst++=',';
      *dst++=hexdigit[(*src>>4)&0xF];
      *dst++=hexdigit[(*src++)&0xF];
    }
  }
  /* copy post-url bit */
  while (*src!='\0') *dst++=*src++;
  /* null-terminate */
  *dst='\0';
}
/* encode string (returns malloc'ed string) */
char *encode(char *base, char *url) {
  char *cp; int is_absolute=0;
  char *retval = malloc(1+strlen(ENCODE_PREFIX)+3*(strlen(base)+strlen(url)));
  CHK_NULL(retval);
  /* search for 'scheme:' prefix */
  for (cp=url; *cp!='\0'; cp++) {
    if (*cp==':') { is_absolute=1; break; }
    if (!isalnum(*cp) && *cp!='+' && *cp!='-' && *cp!='.')
      break; /* not a scheme */
  }
  if (is_absolute && strncasecmp(url, "http:", 5)!=0)
    strcpy(retval, url); /* don't munge; it's not an http scheme. */
  else {
    strcpy(retval, ENCODE_PREFIX);
    if (!is_absolute)
      expand(retval+strlen(retval), base);
    expand(retval+strlen(retval), url);
  }
  return retval;
}

