/* definitions */
%option 8bit caseless fast
%{
#include <stdlib.h>
#include <string.h>
#include "javascript.h"
#include "html.h"
#include "code.h"
#include "err.h"
#define YY_DECL void yylex(char *host, void **dst, int *n, int *size, int script_seen, int is_html)
#define PUTS(str) append(str, dst, n, size)
#define ECHO PUTS(yytext)
#define GROK(str) grok(host, str, dst, n, size)
#define MAKEBUF char buf[strlen(yytext)+1]; strcpy(buf,yytext);

YY_DECL;

void *html_grok(char *host, void *src, int *len) {
  int n=0;
  void *dst = malloc(*len); CHK_NULL(dst);
  yy_scan_bytes(src, *len);
  yylex(host, &dst, &n, len, 0, 1);
  *len = n;
  return dst;
}
void *javascript_grok(char *host, void *src, int *len) {
  int n=0;
  void *dst = malloc(*len); CHK_NULL(dst);
  yy_scan_bytes(src, *len);
  yylex(host, &dst, &n, len, 0, 0);
  *len = n;
  return dst;
}

static void append(char *src, void **dst, int *n, int *size) {
  int slen = strlen(src);
  while (*n+slen > *size) {
    *size *= 2;
    *dst = realloc(*dst, *size); CHK_NULL(*dst);
  }
  memcpy(*dst+*n, src, slen);
  *n += slen;
}

static void grok(char *host, char *_url_, void **dst, int *n, int *size) {
 char *coded, *url=_url_;
#ifdef VERBOSE_DEBUG
 printf("GROKKING: %s\n", url);
#endif
  /* skip over leading whitespace */
  while (*url!='\0' && (*url==' '||*url=='\t'||*url=='\r'||*url=='\n')) url++;
  /* okay, get to work. */
  if (strncasecmp("http:",url,5)==0) {
    coded = encode("", url);
    PUTS(coded);
#ifdef VERBOSE_DEBUG
    printf("HTTP PATH: %s\n", coded);
#endif
    free(coded);
  } else if (url[0]=='/') {
    char buf[strlen(host)+8];
    strcpy(buf, "http://"); strcat(buf, host);
    coded = encode(buf, url);
    PUTS(coded);
#ifdef VERBOSE_DEBUG
    printf("ABS PATH: %s\n", coded);
#endif
    free(coded);
  } else {
    PUTS(_url_);
  }
}

int previous_state=INITIAL;
%}

ALPHA		[A-Za-z]
WS		[\ \t\r\n\v]
NOTWS		[^\ \t\r\n\v]
OPTWS		{WS}*
TAGSTART	"<"|"</"|"<!"
TAGEND		">"
NAME		[A-Za-z][A-Za-z0-9\.\-]*
ANY		.|\n

NOTID		[^A-Za-z0-9_$]

%x tag
%x dstring
%x sstring
%x uri
%x scripttag
%x javascript

%%
	if (is_html) BEGIN(INITIAL); else BEGIN(javascript);

"<"{WS}*script { 
	if (!script_seen) {
	  script_seen=1;
	  PUTS(javascript_code_prefix1);
	  PUTS(host);
	  PUTS(javascript_code_prefix2);
	}
        previous_state=scripttag;
	BEGIN(tag); ECHO;
}
<scripttag>"</"{WS}*script({WS}[^>]*)?">" {
	BEGIN(INITIAL); ECHO; PUTS(javascript_code_suffix);
}
<scripttag,javascript>{NOTID}location{OPTWS}"."{OPTWS}href{OPTWS}"=" {
	char buf[2] = { yytext[0], '\0' };
	PUTS(buf); PUTS("location="); /* we don't like the .href */
}
<scripttag,javascript>{ANY} ECHO;

{TAGSTART}{WS}*{NAME} { BEGIN(tag); ECHO; previous_state=INITIAL; }
<tag>{TAGEND} { BEGIN(previous_state); ECHO; }

<tag>src{OPTWS}={OPTWS} { BEGIN(uri); ECHO; }
<tag>href{OPTWS}={OPTWS} { BEGIN(uri); ECHO; }
<tag>action{OPTWS}={OPTWS} { BEGIN(uri); ECHO; }
<tag>background{OPTWS}={OPTWS} { BEGIN(uri); ECHO; }

"<"{OPTWS}meta{WS}[^>]*http-equiv{OPTWS}={OPTWS}(["]refresh["]|[']refresh[']|refresh){WS}[^>]*content{OPTWS}={OPTWS}["'][^"']*";"{OPTWS}url=[^=;"<>\ ]+ {
	char *equal=strrchr(yytext, '=');
	if (equal!=NULL) {
		char buf[2+equal-yytext];
		memcpy(buf, yytext, 1+equal-yytext);
		buf [1+equal-yytext]='\0';
		PUTS(buf); GROK(equal+1);
	} else ECHO; BEGIN(tag);
}

<tag>"="{OPTWS}[\"] { BEGIN(dstring); ECHO; }
<tag>"="{OPTWS}[\'] { BEGIN(sstring); ECHO; }
<dstring>[\"] { BEGIN(tag); ECHO; }
<sstring>[\'] { BEGIN(tag); ECHO; }
<dstring,sstring>">" { BEGIN(INITIAL); ECHO; /* error recovery? */ }
<dstring,sstring>{ANY} ECHO;

<uri>[\"][^\"\>]*[\"\>] { /* tries to be suave about unterminated quotes */
	char buf[strlen(yytext)];
        strcpy(buf, yytext+1);
	buf[strlen(buf)-1]='\0';
 PUTS("\""); GROK(buf); PUTS(yytext+strlen(yytext)-1); BEGIN(tag);
}
<uri>[\'][^\'\>]*[\'\>] { /* tries to be suave about unterminated quotes */
	char buf[strlen(yytext)];
        strcpy(buf, yytext+1);
	buf[strlen(buf)-1]='\0';
 PUTS("\'"); GROK(buf); PUTS(yytext+strlen(yytext)-1); BEGIN(tag);
}
<uri>{NAME}|([^\"\'][^<>\ \t\v\f\r\n]*) {
 /* regexp used to be: ([A-Za-z0-9_/\.:#\-\?\&]+) */
 /* some sites doesn't use quotes properly */
 PUTS("\""); GROK(yytext); PUTS("\""); BEGIN(tag);
}
<uri>{ANY} { ECHO; BEGIN(tag); /* error */ }

{ANY} ECHO;
<<EOF>> return;
%%
