/* HTTP request parser. */
#include <openssl/ssl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include "linereader.h"
#include "writer.h"
#include "http.h"
#include "html.h"
#include "code.h"
#include "err.h"

#define NL "\r\n"

#define SSL_write SSL_write_fully
#define write write_fully

#define prefixcmp(str, line) strncasecmp(line, str, strlen(str))
#define writeString(fd, str) write(fd, str, strlen(str))
#define SSL_writeString(ssl, str) SSL_write(ssl, str, strlen(str))

int generate_not_found(SSL *ssl, LR_input_t * ssl_in, char *url, int is_head);
int generate_redirect(SSL *ssl, LR_input_t * ssl_in, char *host, int is_head);

int proxy(SSL *ssl, LR_input_t * ssl_in, char *host, int port, char *doc, 
	   char *request);

static int my_ssl_read(void *where, void *buf, int size) {
  return SSL_read((SSL*)where, buf, size);
}

static int my_fd_read(void *where, void *buf, int size) {
  return read(*((int *)where), buf, size);
}

void http_parse(SSL *ssl, const char *from) {
  LR_input_t ssl_in = LR_init(my_ssl_read, ssl, "");
  int close_now = 0;

  char *line;
  while (!close_now && (line = LR_readline(&ssl_in)) != NULL) {
    int is_head = (0==prefixcmp(line, "HEAD "));
    /* line is request-line */
    if (strstr(line, " HTTP/")!=NULL) {
      char *skip, *url;
      skip=strchr(line,' ')+1; if (*skip=='/') skip++;
      url = decode(skip);
      if (from!=NULL)
	printf("%s %s\n", from, url);
      if (prefixcmp("http:", url)!=0 || strstr(url," HTTP/")==NULL) {
	/* error document */
	close_now = generate_not_found(ssl, &ssl_in, url, is_head);
      } else {
	char *host, *port, *doc, *http=strstr(url," HTTP/");
	int p, trailing_slash=(NULL!=strstr(url, "/ HTTP/"));
	/* split HTTP/1.x modifier */
	*http++='\0';
	/* skip over slashes after http: */
	for (host=url+5; *host=='/'; ) host++;
	/* find slash-termination of document name */
	doc=strchr(host,'/');
	if (doc==NULL) doc=""; else while (*doc=='/') *doc++='\0';
	/* separate out port number */
	port = strchr(host,':');
	if (port==NULL || port>=doc) p=80;
	else { *port++='\0'; p=atoi(port); }
	/* if there were any spaces in the doc (before the HTTP), kill 'em */
	if (strchr(doc, ' ')!=NULL) *(strchr(doc, ' '))='\0';
	/* make the new request-uri */
	strcpy(skip, doc);
	strcat(skip, " ");
	strcat(skip, http);
#ifdef VERBOSE_DEBUG
	printf("FROM SSL0: %s (%s:%d)\n", line, host, p);
#endif
	if (strstr(host, ".anonymise.com")!=NULL ||
	    strstr(host, "206.253.222.")!=NULL ) {
	  close_now = generate_redirect(ssl, &ssl_in,
					"https://lm.lcs.mit.edu/scum.html",
					is_head);
	} else if (doc[0]=='\0' && !trailing_slash) {
	  char buf[strlen(host)+10];
	  sprintf(buf, "http://%s/", host);
	  close_now = generate_redirect(ssl, &ssl_in, buf, is_head);
	} else
	  close_now = proxy(ssl, &ssl_in, host, p, doc, line);
      }
      free(url);
    } else if (strlen(line)==0) continue; /* standard says to skip. */
    else printf("UNKNOWN: %s\n", line);
  }
  LR_free(&ssl_in);
}

int proxy(SSL *ssl, LR_input_t * ssl_in, char *host, int port, char *doc,
	   char *request) {
  char *line;
  int sd, err;
  struct sockaddr_in sa;
  struct hostent *hinfo;
  LR_input_t sd_in = LR_init(my_fd_read, &sd, "");
  int content_length;
  int content_base;
  int content_text_html, content_javascript;
  int transfer_encoding_chunked;
  int is_head_request;
  int connection_close;
  int saw_content_type;
  int retval = 0;

  sd = socket(AF_INET, SOCK_STREAM, 0); CHK_ERR(sd, "proxy socket");
  memset(&sa, 0, sizeof(sa));
  sa.sin_family	     = AF_INET;
  sa.sin_port        = htons(port);
  /** First try looking up as aaa.bbb.ccc.ddd */
  if (-1 == (sa.sin_addr.s_addr = inet_addr(host))) {
    /* OK, look up by name. */
    hinfo = gethostbyname(host); CHK_NULL(hinfo);
    memcpy(&sa.sin_addr, hinfo->h_addr, hinfo->h_length);
  }

  err = connect(sd, (struct sockaddr*) &sa, sizeof(sa));
  CHK_ERR(err,"proxy connect");
  
  is_head_request = (strncmp("HEAD ", request, 5)==0);

  writeString(sd, request);
  writeString(sd, NL);
  content_length=-1;
  while(strlen(line=LR_readline(ssl_in))>0) {
    /** FIXME: look for Content-Length or Transfer-Encoding */
#ifdef VERBOSE_DEBUG
    printf("FROM SSL1: %s\n", line);
#endif
    if (prefixcmp("Referer: ", line)==0)
      continue; /* suppress referer information. */
    if (prefixcmp("Connection: close", line)==0)
      retval=1; /* client is telling us to close connection after this */
    if (prefixcmp("Connection: ", line)==0)
      continue; /* make connections non-persistent. */
    if (prefixcmp("Content-Length: ",line)==0)
      content_length=atoi(line+16);
    if (prefixcmp("Host: ", line)==0) {
      char buf[strlen(host)+26];
      if (port==80) sprintf(buf, "Host: %s", host);
      else sprintf(buf, "Host: %s:%d", host, port);
      writeString(sd, buf);
    } else {
      writeString(sd, line);
    }
    writeString(sd, NL);
  }
  /** Make connections non-persistent. */
  writeString(sd, "Connection: close" NL);
  writeString(sd, NL);
  /* dump content, if present */
  if (content_length>0) { int i;
    char *buf = LR_rawread(ssl_in, &content_length); CHK_NULL(buf);
    write(sd, buf, content_length);
  }

  content_length=-1; content_base = 0;
  content_text_html = 0; content_javascript = 0;
  transfer_encoding_chunked = 0; connection_close = 0;
  saw_content_type = 0;
  while(strlen(line=LR_readline(&sd_in))>0) {
    char *replacement = NULL;
    /** FIXME: look for Transfer-Encoding */
    /** FIXME: Content-Location URI */
#ifdef VERBOSE_DEBUG
    printf("FROM SOCK: %s\n", line);
#endif
    if (prefixcmp("Content-Length: ",line)==0) {
      content_length=atoi(line+16);
      continue; /* suppress */
    }
    if (prefixcmp("Connection: close", line)==0)
      connection_close=1;
    if (prefixcmp("Content-Type: ",line)==0) {
#ifdef VERBOSE_DEBUG
      printf("CONTENT-TYPE: %s\n", line+14);
#endif
      saw_content_type = 1;
    }
    if (prefixcmp("Content-Type: application/x-javascript",line)==0)
      content_javascript = 1;
    if (prefixcmp("Content-Type: text/html",line)==0)
      content_text_html = 1;
    if (prefixcmp("Transfer-Encoding: chunked",line)==0)
      transfer_encoding_chunked=1;
    if (prefixcmp("Location: ",line)==0) {
      char buf[strlen(host)+30], *url;
      if (port==80) sprintf(buf, "http://%s/", host);
      else sprintf(buf, "http://%s:%d/", host, port);
      url = encode(buf,line+10);
      err = SSL_write(ssl, line, 10); CHK_SSL(err);
      err = SSL_writeString(ssl, url); CHK_SSL(err);
      free(url);
    } else if (prefixcmp("Content-Base: ",line)==0) {
      char buf[strlen(host)+30], *url;
      if (port==80) sprintf(buf, "http://%s/", host);
      else sprintf(buf, "http://%s:%d/", host, port);
      url = encode(buf,line+14);
      err = SSL_write(ssl, line, 14); CHK_SSL(err);
      err = SSL_writeString(ssl, url); CHK_SSL(err);
      free(url);
      content_base=1;
    } else {
      err = SSL_writeString(ssl, line); CHK_SSL(err);
    }
    err = SSL_writeString(ssl, NL); CHK_SSL(err);
  }
  if (!content_base) { /* hmm. netscape seems to ignore this field. */
    char buf[strlen(host)+strlen(doc)+30], *url;
    if (port==80) sprintf(buf, "http://%s/%s", host, doc);
    else sprintf(buf, "http://%s:%d/%s", host, port, doc);
    url = encode("",buf);
    err = SSL_writeString(ssl, "Content-Base: "); CHK_SSL(err);
    err = SSL_writeString(ssl, url); CHK_SSL(err);
    err = SSL_writeString(ssl, NL); CHK_SSL(err);
    free(url);
  }
  /* Sun's "JavaWebServer/1.1.3" on developer.javasoft.com doesn't
   * bother to send a content-type field across for text/html.
   * is this legal? */
  if (!saw_content_type) { /* see if we can fill in a bit here */
    int dlen = strlen(doc);
    if (dlen>=3 && strcasecmp(doc+dlen-3, ".js")==0) {
      content_javascript = 1; saw_content_type = 1;
      err = SSL_writeString(ssl, "Content-Type: application/x-javascript");
      CHK_SSL(err);
      err = SSL_writeString(ssl, NL); CHK_SSL(err);
    } else if (dlen>=4 && strcasecmp(doc+dlen-4, "html")==0) {
      content_text_html = 1; saw_content_type = 1;
      err = SSL_writeString(ssl, "Content-Type: text/html"); CHK_SSL(err);
      err = SSL_writeString(ssl, NL); CHK_SSL(err);
    }
  }
  if (is_head_request) {
    err = SSL_writeString(ssl, NL); CHK_SSL(err);
  } else/*if (content_length>=0 || connection_close)*/{/*allow for weird CGIs*/
    void *content=NULL, *buf;
    if (transfer_encoding_chunked) {
      int chunk_size;
      content_length = 0;
      line=LR_readline(&sd_in);
      chunk_size = strtol(line, NULL, 16);
      while (chunk_size > 0) {
	content=realloc(content, content_length+chunk_size); CHK_NULL(content);
	buf = LR_rawread(&sd_in, &chunk_size); CHK_NULL(buf);
	memcpy(content+content_length, buf, chunk_size);
	content_length+=chunk_size;
	line = LR_readline(&sd_in);
	if (strlen(line)!=0) printf("SOMETHING'S WRONG!\n");
	line = LR_readline(&sd_in);
	chunk_size = strtol(line, NULL, 16);
      }
    } else if (content_length<0) { /* weird CGIs */
      /* Safe to read 'till close because we sent "Connection: close" */
      int size = 16 * 1024;
      content = malloc(size);
      for (content_length=0; ; ) {
	int n = size-content_length;
	void * b = LR_rawread(&sd_in, &n);
	if (b==NULL || n<=0) break;
	memcpy(content+content_length, b, n);
	content_length+=n;
	if (size < 2*content_length) {
	  size = 2*content_length;
	  content = realloc(content, size);
	}
      }
#ifdef VERBOSE_DEBUG
      printf("CGI CONTENT LENGTH: %d\n", content_length);
#endif
    } else {
      buf = LR_rawread(&sd_in, &content_length); CHK_NULL(buf);
      content = malloc(content_length); CHK_NULL(content);
      memcpy(content, buf, content_length);
    }
    /* OKAY, last resort for absent Content-Type field: grep for <html> */
    if (!saw_content_type) {
      void *ptr = content, *end = content+content_length;
      for (ptr=memchr(content,'<',end-ptr); ptr!=NULL && end-ptr>=6;
	   ptr=memchr(ptr+1,'<',end-ptr-1)) {
	/* FIXME: skip over leading/trailing white space */
	if (prefixcmp("<html>", ptr)==0) { /* whoopee! */
	  content_text_html = 1; saw_content_type = 1;
	  err = SSL_writeString(ssl, "Content-Type: text/html"); CHK_SSL(err);
	  err = SSL_writeString(ssl, NL); CHK_SSL(err);
	  break;
	}
      }
    }
    /* grok the file if it's javascript or html */
    if (content_text_html) {
      void *grokked = html_grok(host, content, &content_length);
      free(content);
      content = grokked;
    } else if (content_javascript) {
      void *grokked = javascript_grok(host, content, &content_length);
      free(content);
      content = grokked;
    }
    /* output the content-length */
    if (!transfer_encoding_chunked) {
      char buf[30];
      sprintf(buf, "Content-Length: %d" NL, content_length);
      err = SSL_writeString(ssl, buf);
    }
    /* terminate the header section */
    err = SSL_writeString(ssl, NL); CHK_SSL(err);
    /* output the chunk header */
    if (transfer_encoding_chunked) {
      char buf[25];
      sprintf(buf, "%d" NL, content_length);
      err = SSL_writeString(ssl, buf); CHK_SSL(err);
    }
    err = SSL_write(ssl, content, content_length); CHK_SSL(err);
    free(content);
    if (transfer_encoding_chunked) {
      SSL_writeString(ssl, NL "0" NL);
      while(strlen(line=LR_readline(&sd_in))>0) {
	/* FIXME: we're not processing these lines */
	err = SSL_writeString(ssl, line); CHK_SSL(err);
	err = SSL_writeString(ssl, NL); CHK_SSL(err);
      }
      err = SSL_writeString(ssl, NL); CHK_SSL(err);
    }
  }
  LR_free(&sd_in);
  close(sd);
  return retval;
}

int generate_not_found(SSL *ssl, LR_input_t * ssl_in, char *url, int is_head) {
  char *line;
  while (strlen(line=LR_readline(ssl_in))>0)
    /* chew up */;
  SSL_writeString(ssl, 
		  "HTTP/1.1 404 Not Found" NL
		  "Server: SecureProxy" NL
		  "Connection: close" NL
		  "Content-Type: text/html" NL
		  NL);
  if (is_head) return 1;
  SSL_writeString(ssl, 
		  "<HTML><HEAD><TITLE>Invalid URL: ");
  SSL_writeString(ssl, url);
  SSL_writeString(ssl,
		  "</TITLE></HEAD>" NL
		  "<BODY>" NL
		  "<H1 ALIGN=CENTER>Invalid URL</H1>" NL
		  "<HR>" NL
		  "<P>Could not parse the requested url: ");
  SSL_writeString(ssl, url);
  SSL_writeString(ssl, "</P>" NL
		  "<HR>" NL
		  "<TABLE WIDTH=\"100%\" BORDER=0><TR><TD ALIGN=RIGHT>" NL
		  "<ADDRESS>cananian@mit.edu</ADDRESS>" NL
		  "</TD></TR></TABLE>" NL
		  "</BODY>" NL
		  "</HTML>" NL);
  return 1;
}

int generate_redirect(SSL *ssl, LR_input_t * ssl_in, char *url, int is_head) {
  char *enc, *line;

  while (strlen(line=LR_readline(ssl_in))>0)
    /* chew up */;

  enc = encode("", url);

  SSL_writeString(ssl, 
		  "HTTP/1.1 302 Found" NL
		  "Server: SecureProxy" NL
		  "Location: ");
  SSL_writeString(ssl, enc);
  SSL_writeString(ssl, NL
		  "Connection: close" NL
		  "Content-Type: text/html" NL
		  NL);
  if (is_head) { free(enc); return 1; }
  SSL_writeString(ssl,
		  "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">" NL
		  "<HTML><HEAD>" NL
		  "<TITLE>302 Found</TITLE>" NL
		  "</HEAD><BODY>" NL
		  "<H1>Found</H1>" NL
		  "The document has moved <A HREF=\"");
  SSL_writeString(ssl, enc);
  SSL_writeString(ssl,
		  "\">here</A>.<P>" NL
		  "</BODY></HTML>" NL);
  free(enc);
  return 1;
}
